/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: asgraph.h                                    */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#ifndef ASGRAPH_H
#define ASGRAPH_H

#define MAX_COMMENT_CHARS 512

#include "graph.h"
#include "iptools.h"
#include <stdio.h>
#include <string.h>

// *************************************
// ************** AS graph *************
// *************************************

typedef unsigned short int ASnode;

/* The following type simplifies references to differential attributes
   in edge weights */
enum cmp_graph {CMPGRAPH_FIRST = 0, CMPGRAPH_SECOND = 1};

// AS graph's edges are labelled with miscellaneous information
struct ASEdgeWeight {
   ASEdgeWeight();                           // Default constructor
   ASEdgeWeight (const ASEdgeWeight &w);     // Copy constructor
   ASEdgeWeight &operator=
         (const ASEdgeWeight &w);            // Assignment operator

   /* When the graph a differential one (i.e., it was built by comparing two
      other graphs), it is possible to see the properties of an edge as they
      appeared on the source graphs, by simply checking CMPGRAPH_FIRST and
      CMPGRAPH_SECOND values; when it is not a differential one, the only
      significant values are those stored inside the first element (CMPGRAPH_FIRST
      of each array) */
   bool oriented;                   /* This field is set to false when edge
                                       direction is meaningless (i.e., edge
                                       has a direction only because otherwise
                                       it would not exist...) */
   unsigned long int
                  pathCovering[2];  // Number of paths that traverse current edge
   unsigned long long int
                  IPCovering[2];    /* Number of addresses that were announced over
                                       current edge */
   IPPrefix lastPrefix;             /* This is a temporary field that is used when
                                       updating graph with IP address covering */

   // These variables are only used when the graph is a differential one
   bool diffOriented[2];            /* Differential graphs already use `oriented' flag
                                       to mark their own edges; to know whether an
                                       edge was oriented on the compared graphs, use
                                       this variable instead */
   edge_dir orientation[2];         /* This stores the direction of the edge
                                       as it appears in the compared graphs (same
                                       note as above) */
   unsigned short int graphs,       // Number of compared graphs
                      incoming,     // Number of times this edge's direction is DIR_IN
                      outgoing,     // Number of times this edge's direction is DIR_OUT
                      none,         // Number of times this edge's direction is DIR_NONE
                      undir,        // Number of times this edge appears unoriented
                      snapshots;    /* This variable is temporarily used when computing
                                       the mean duration of an edge orientation */
   float stability;                 /* This stores the mean duration of an orientation
                                       over current edge, in terms of number of
                                       snapshots */
   unsigned short int changes;      /* This field contains the number of times that
                                       current edge's orientation has changed */
};

class ASGraph: public Graph<ASnode, ASEdgeWeight, ASnode> {
   private:
      bool differential;                        /* When this flag is set to true,
                                                   graph is a differential graph
                                                   (i.e., it was obtained by computing
                                                   the differences between two or
                                                   more 'conventional' graphs) */
      char error[513];                          /* This variable stores a short
                                                   explanation for error conditions,
                                                   when they occur; 513 includes 512
                                                   characters plus terminating '\0' */
      List<char*> sourceFile;                   /* A graph is always labelled with the
                                                   name of the files it was generated
                                                   from */
      char description[MAX_COMMENT_CHARS + 1];  /* A short description of what the graph
                                                   represents */
      int dirToVal(edge_dir);                   /* This funcion converts an edge direction
                                                   (IN, OUT, NONE) into a printable integer
                                                   value (-1, 1, 0) */
      edge_dir valToDir(int);                   /* This function converts an integer value
                                                   (-1, 1, 0) into an edge direction
                                                   (IN, OUT, NONE) */
      /* The following two constants store character sequences that should
         identify the type of the graph when written to file */
      static const char graphMarker[4];
      static const char diffGraphMarker[4];

   public:
      ASGraph();                       // Default constructor
      ~ASGraph();                      // Destructor
      bool writeToFile(FILE*);         /* Dumps all the AS graph contents to an
                                          ASCII file; use `readFromFile' to
                                          get them back; the only parameter is
                                          a pointer to an open file strem;
                                          this method returns true if operation
                                          is successful; false otherwise */
      bool readFromFile(FILE*);        /* Builds a graph from the data that have
                                          previously been generated by `writeToFile';
                                          the only parameter is a pointer to an
                                          open file stream; returns true if
                                          operation is successful; false otherwise */
      char *errReason();               /* Returns a string that describes the
                                          reason of last file error; this method
                                          should be invoked after that either one
                                          of `writeToFile' or `readFromFile'
                                          returned false */
      List<char*> &getSourceFiles();   /* Returns a list of the names of the files
                                          the graph was generated from */
      void setSourceFiles
                  (char**, bool);      /* Sets the graph's source file names; if
                                          the second argument is true, current list
                                          of source files is also cleared */
      void ASGraph::setSourceFiles
               (List<char*>&, bool);   /* The same as above, but input is a list
                                          (instead of an array) of file names */
      char *getComment();              /* Returns the comment associated with the
                                          graph; an empty string if it has no comment */
      void setComment(char*);          /* Sets the graph comment to a given value;
                                          if argument is longer than MAX_COMMENT_CHARS
                                          characters, it will be truncated */
      bool &isDifferential();          /* This function allows to read and/or
                                          change the state of the flag that
                                          marks the graph as a differential one;
                                          it must be set to true whenever
                                          differential information should be
                                          saved to file */
};

// ************** AS graph *************

#endif
