/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: fileparsers.h                                */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#ifndef FILEPARSERS_H
#define FILEPARSERS_H

#include "parser.h"
#include "iptools.h"
#include "list.h"

// ******************************************
// ************ BGP dump parser *************
// ******************************************

enum fields {FIELD_NETWORK = 0,           // These are all the fields in each line
            FIELD_NEXTHOP = 1,            // of the BGP dump
            FIELD_METRIC = 2,
            FIELD_LOCPRF = 3,
            FIELD_WEIGHT = 4,
            FIELD_ASPATH = 5,
            FIELD_MORE = 6};
enum as_type {AS_PRIVATE = 0x1,           // These flags identify the role of an AS
              AS_SET = 0x2};              // inside a path
enum route_stat {ROUTE_SUPPRESSED = 0x1,  // Each of these is a possible status of
                 ROUTE_DAMPED = 0x2,      // a route
                 ROUTE_HISTORY = 0x4,
                 ROUTE_VALID = 0x8,
                 ROUTE_BEST = 0x10,
                 ROUTE_INTERNAL = 0x20,
                 ROUTE_STALE = 0x40};
enum route_origin {ORIGIN_IGP = 0x1,      // These are the possible route origins
                   ORIGIN_EGP = 0x2,
                   ORIGIN_INCOMPLETE = 0x4};

class bgpParser: public Parser {
   private:
      /* The following are field captions (i.e. the header names of each field) and
         status characters. They are initialized right after the end of the class
         declaration */
      static const char fieldCaptions[7][9];
      static const char routeStatCodes[7];
      static const char routeOriginCodes[3];

      // The following are characters used as special delimiters
      static const char privateAS[2];
      static const char ASset[2];
      IPAddress *lastNetwork;                   // These are used to remember the announced
      IPNetmask *lastNetmask;                   // prefix of the previous line

   public:
      struct ASItem {
         unsigned short int ASnum;              // AS number
         unsigned char flags;                   /* AS properties; bit mask is used as follows:
                                                   00000000    "normal" AS
                                                   00000001    private AS
                                                   00000010    AS inside a set */
      };
      struct parsed_data {                      /* This structure holds data obtained from
                                                   last parsed line */
         IPAddress network;                     // Network (IP address)
         IPNetmask *netmask;                    // Netmask (if any)
         IPAddress nextHop;                     // Next hop
         unsigned char route_flags;             /* Route properties; bit mask is used as follows:
                                                   00000000    "normal" route
                                                   00000001    suppressed route
                                                   00000010    damped route
                                                   00000100    "history" route
                                                   00001000    valid route
                                                   00010000    best route
                                                   00100000    internal route
                                                   01000000    "stale" route
                                                   N.B.: Direct use of the enum type route_stat is
                                                   forbidden because route_flags can be more
                                                   than one */
         unsigned char path_flags;              /* AS path properties; bit mask is used as follows:
                                                   00000000    route learned via IGP
                                                   00000001    route learned via EGP
                                                   00000010    incomplete path
                                                   See the note for route_flags above */
         unsigned long long int *metric;        // Route metric
         unsigned int *locpref;                 // Route local preference
         unsigned int *weight;                  // Route weight
         List<ASItem> ASpath;                   // AS path
      };
      bgpParser();                              // Default constructor
      ~bgpParser();                             // Destructor
      parsed_data parsedData;                   // This stores (and makes public) parsed data
      parse_stat parseLine();                   /* Attempts to parse current line; before
                                                   doing that, you must use nextLine in
                                                   order to move over the first parsable
                                                   line; if you forget doing so, a PARSE_NOLINE
                                                   error will be returned */
      parsefile_stat nextLine();                // Moves to the next parsable line in the file
};

// ************ BGP dump parser *************






// ******************************************
// ********* bgpSat output parser ***********
// ******************************************

class satParser: public Parser {
   private:
      static const char peerStr[18],            /* These constant hold the strings that
                                                   identify `candidate peering' lines... */
                        badStr[10];             // ...and `bad path' ones
   public:
      struct parsed_data {                      /* This structure holds data obtained from
                                                   last parsed line */
         unsigned short int from, to;           // A directed edge
         double *straight, *negated;            // Edge attributes
         bool reversed;                         /* This field is true when current edge has
                                                   been reversed while computing orientation */
         bool peering;                          /* This is set to true when last parsed
                                                   line is a "candidate peering" one... */
         bool bad;                              /* ...and this one when it is a "bad path"
                                                   one */
         List<unsigned short int> ASpath;       /* This list is used when returning a bad
                                                   path */
      };
      satParser();                              // Default constructor
      ~satParser();                             // Destructor
      parsed_data parsedData;                   // This stores (and makes public) parsed data
      parse_stat parseLine();                   /* Attempts to parse current line; before
                                                   doing that, you must use nextLine in
                                                   order to move over the first parsable
                                                   line; if you forget doing so, a PARSE_NOLINE
                                                   error will be returned */
      parsefile_stat nextLine();                // Moves to the next parsable line in the file
};

// ********* bgpSat output parser ***********






// ******************************************
// ********* Agarwal graph parser ***********
// ******************************************

class vantageParser: public Parser {
   public:
      struct parsed_data {                      /* This structure holds data obtained from
                                                   last parsed line */
         unsigned short int from, to;           /* These two values represent the two ends
                                                   of an edge */
         bool peering;                          /* This flag is true if current edge is a
                                                   peering edge (i.e., edge direction is
                                                   meaningless - from and to can be arbitrarily
                                                   exchanged) */
      };
      parsed_data parsedData;                   // This stores (and makes public) parsed data
      parse_stat parseLine();                   /* Attempts to parse current line; before
                                                   doing that, you must use nextLine in
                                                   order to move over the first parsable
                                                   line; if you forget doing so, a PARSE_NOLINE
                                                   error will be returned */
      parsefile_stat nextLine();                // Moves to the next parsable line in the file
};

// ********* Agarwal graph parser ***********





// ******************************************
// ********** AS Path file parser ***********
// ******************************************

class pathParser: public Parser {
   public:
      struct parsed_data {                      /* This structure holds data obtained from
                                                   last parsed line */
         List<unsigned short int> ASpath;       // A list of the AS number in the path
      };
      parsed_data parsedData;                   // This stores (and makes public) parsed data
      parse_stat parseLine();                   /* Attempts to parse current line; before
                                                   doing that, you must use nextLine in
                                                   order to move over the first parsable
                                                   line; if you forget doing so, a PARSE_NOLINE
                                                   error will be returned */
      parsefile_stat nextLine();                // Moves to the next parsable line in the file
};

// ********** AS Path file parser ***********







// ******************************************
// ******* Unknown edges file parser ********
// ******************************************

class unknownParser: public Parser {
   public:
      struct parsed_data {                      /* This structure holds data obtained from
                                                   last parsed line */
         unsigned short int from, to;           // The two ends of last parsed edge
      };
      parsed_data parsedData;                   // This stores (and makes public) parsed data
      parse_stat parseLine();                   /* Attempts to parse current line; before
                                                   doing that, you must use nextLine in
                                                   order to move over the first parsable
                                                   line; if you forget doing so, a PARSE_NOLINE
                                                   error will be returned */
      parsefile_stat nextLine();                // Moves to the next parsable line in the file
};

// ******* Unknown edges file parser ********

#endif
