/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: list.h                                       */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#ifndef LIST_H
#define LIST_H

#include <stdlib.h>

// **************************************
// ***************  List  ***************
// **************************************

template <class elType> class List {
   private:
      struct list_element {
         elType contents;                 // Contents of the element
         list_element *previous, *next;   // Pointers to the previous and next element
      };
      list_element *top,                  // Pointer to the first element
                   *cur,                  // Pointer to the current element
                   *bottom;               // Pointer to the last element
      unsigned long long int length;      // Traces list length
   public:
      List();                             // Default constructor
      ~List();                            // Destructor
      List(const List<elType>&);          // Copy constructor
      List<elType>& operator=(            // Assignment operator overloading
                  const List<elType>&);
      bool isEmpty();                     // Returns true if the list is empty
      void push(const elType&);           // Add an element before the beginning of the list
      void enqueue(const elType&);        // Add an element after the end of the list
      void insertA(const elType&);        // Add an element after the current one
      void insertB(const elType&);        // Add an element before the current one
      elType first();                     /* Returns the first element of the list (top of the
                                             stack and of the queue); a non-significant value if
                                             the list is empty */
      bool isFirst();                     /* Checks whether current element is the first one;
                                             returns true for empty lists */
      void goToFirst();                   /* Goes to the first element of the list (top of the
                                             stack and of the queue) */
      bool prev();                        /* Makes the previous element in the list the
                                             current element; if the old current was the
                                             first element, the function returns false, and
                                             no change takes place; otherwise, it returns
                                             true */
      elType current();                   /* Returns the current element of the list; a
                                             non-significant value if the list is empty */
      bool next();                        /* Makes the next element in the list the
                                             current element; the returned value is the
                                             same as prev */
      void goToLast();                    /* Goes to the last element of the list (bottom of
                                             the stack, tail of the queue) */
      bool isLast();                      /* Checks whether current element is the last one;
                                             returns true for empty lists */
      elType last();                      /* Returns the last element of the list (bottom of
                                             the stack, tail of the queue); a non-significant
                                             value if the list is empty  */
      elType pop();                       /* Get the element at the top of the list, and
                                             delete it; returns a non-significant value if
                                             the list is empty */
      elType dequeue();                   // Same as pop
      unsigned long long int len();       // Returns the length of the list
      void removeCurrent();               /* Removes the current element from the list; does
                                             nothing if the list is empty; current element is
                                             moved forward (if possible) */
      void sort();                        /* Sorts list elements, using the merge sort algorithm;
                                             sorting order is ascending (a reversed sort can
                                             easily be obtained by exploring the list in reversed
                                             order); ordering exploits '>' operator, so
                                             redefine it in order to obtain a different
                                             arrangement */
};
/* Examples follow:
   initially      l = {}
   after l.push(1)      l = {*1} (*=current element)
   after l.push(2)      l = {2,*1}
   after l.enqueue(0)   l = {2,*1,0}
   after l.goToFirst()  l = {*2,1,0}
*/




/* Warning: using this code assumes that your C++ compiler supports
   partial expression evaluation (e.g. In (A || B) B is not evaluated
   at all if A is true */

// Default constructor
template <class elType> List<elType>::List() {
   top = cur = bottom = NULL;
   length = 0ULL;
}

// Destructor
template <class elType> List<elType>::~List() {
   list_element *temporaryTop;
   while (top!=NULL) {
      temporaryTop = top;
      top = top->next;
      delete temporaryTop;
   }
}

// Copy constructor
template <class elType> List<elType>::List(const List<elType>& list) {
   list_element *currentEl;
   top = bottom = cur = NULL;
   length = 0ULL;
   if (list.top!=NULL) {
      /* The following copying instructions do not use standard methods (e.g. list.next()),
         in order not to alter current pointer setting */
      for (currentEl = list.top; currentEl!=NULL; currentEl = currentEl->next) {
         enqueue(currentEl->contents);
         if (list.cur==currentEl) { cur = bottom; }
      }
   }
}

// Assignment operator overloading
template <class elType> List<elType>& List<elType>::operator=(const List<elType>& list) {
   list_element *destEl, *sourceEl, *temp;
   bool currentAssigned = false;
   destEl = top;
   sourceEl = list.top;
   // As far as the two lists have the same size, just copy their contents
   while (destEl!=NULL && sourceEl!=NULL) {
      destEl->contents = sourceEl->contents;
      if (list.cur==sourceEl) { cur = destEl; currentAssigned = true; }
      destEl = destEl->next;
      sourceEl = sourceEl->next;
   }
   /* The following is executed if the source list terminates before the destination list
      (this object) */
   if (destEl!=NULL) {
      // This "disconnects" the trail from the preceding elements (if any)
      if (destEl->previous!=NULL) {
         destEl->previous->next = NULL;
      }
      else {
         // There were no previous elements in the list
         top = NULL;
      }
      bottom = destEl->previous;
      if (!currentAssigned) {
         // The pointer to the current element has not been processed yet
         cur = top;
      }
   }
   // Delete trailing elements (if any)
   while (destEl!=NULL) {
      temp = destEl;
      destEl = destEl->next;
      delete temp;
   }
   /* The following is executed if the destination list (this object) terminates before the
      source list */
   while (sourceEl!=NULL) {
      enqueue(sourceEl->contents);
      if (list.cur==sourceEl) { cur = bottom; }
      sourceEl = sourceEl->next;
   }
   length = list.length;
   return *this;
}

// Returns true if the list is empty
template <class elType> bool List<elType>::isEmpty() {
   return (cur==NULL);  /* There is no need to check the other pointers:
                           if just any of them is suddenly null, this means
                           that no more elements exist in the list */
}

// Add an element before the beginning of the list
template <class elType> void List<elType>::push(const elType& stuff) {
   list_element *newElement;
   newElement = new list_element;
   newElement->contents = stuff;
   newElement->next = top;
   newElement->previous = NULL;
   if (top!=NULL) {
      // The list wasn't empty
      top->previous = newElement;
   }
   top = newElement;
   if (bottom==NULL) {
      // The list was previously empty
      bottom = top;
      cur = top;
   }
   length++;
}

// Add an element after the end of the list
template <class elType> void List<elType>::enqueue(const elType& stuff) {
   list_element *newElement;
   newElement = new list_element;
   newElement->contents = stuff;
   newElement->next = NULL;
   newElement->previous = bottom;
   if (bottom!=NULL) {
      // The list wasn't empty
      bottom->next = newElement;
   }
   bottom = newElement;
   if (top==NULL) {
      // The list was previously empty
      top = bottom;
      cur = bottom;
   }
   length++;
}

// Add an element after the current one
template <class elType> void List<elType>::insertA(const elType& stuff) {
   list_element *newElement;
   newElement = new list_element;
   newElement->contents = stuff;
   newElement->previous = cur;
   if (cur!=NULL) {
      newElement->next = cur->next;
      if (cur->next == NULL) {
         // Current element is also the last one
         bottom = newElement;
      }
      else {
         cur->next->previous = newElement;
      }
      cur->next = newElement;
   }
   else {
      // The list was previously empty
      newElement->next = NULL;
      cur = top = bottom = newElement;
   }
   length++;
}

// Add an element before the current one
template <class elType> void List<elType>::insertB(const elType& stuff) {
   list_element *newElement;
   newElement = new list_element;
   newElement->contents = stuff;
   newElement->next = cur;
   if (cur!=NULL) {
      newElement->previous = cur->previous;
      if (cur->previous == NULL) {
         // Current element is also the first one
         top = newElement;
      }
      else {
         cur->previous->next = newElement;
      }
      cur->previous = newElement;
   }
   else {
      // The list was previously empty
      newElement->previous = NULL;
      cur = top = bottom = newElement;
   }
   length++;
}

/* Returns the first element of the list (top of the stack and of the queue);
   a non-significant value if the list is empty */
template <class elType> elType List<elType>::first() {
   if (top!=NULL) {
      return (top->contents);
   }
   else {
      elType dummy;
      return (dummy);
   }
}

// Checks whether current element is the first one; returns true for empty lists
template <class elType> bool List<elType>::isFirst() {
   return (cur==top);
}

// Goes to the first element of the list (top of the stack and of the queue)
template <class elType> void List<elType>::goToFirst() {
   cur = top;
}

/* Makes the previous element in the list the
   current element; if the old current was the
   first element, the function returns false, and
   no change takes place; otherwise, it returns
   true */
template <class elType> bool List<elType>::prev() {
   if ((cur==NULL) || (cur->previous==NULL)) {
      return (false);
   }
   cur = cur->previous;
   return (true);
}

/* Returns the current element of the list; a
   non-significant value if the list is empty */
template <class elType> elType List<elType>::current() {
   if (cur!=NULL) {
      return (cur->contents);
   }
   else {
      elType dummy;
      return (dummy);
   }
}

/* Makes the next element in the list the
   current element; the returned value is the
   same as prev */
template <class elType> bool List<elType>::next() {
   if ((cur==NULL) || (cur->next==NULL)) {
      return (false);
   }
   cur = cur->next;
   return (true);
}

// Goes to the last element of the list (bottom of the stack, tail of the queue)
template <class elType> void List<elType>::goToLast() {
   cur = bottom;
}

// Checks whether current element is the last one; returns true for empty lists
template <class elType> bool List<elType>::isLast() {
   return (cur==bottom);
}

/* Returns the last element of the list (bottom of the stack, tail of the queue);
   a non-significant value if the list is empty  */
template <class elType> elType List<elType>::last() {
   if (bottom!=NULL) {
      return (bottom->contents);
   }
   else {
      elType dummy;
      return (dummy);
   }
}

/* Get the element at the top of the list, and delete it; returns a
   non-significant value if the list is empty */
template <class elType> elType List<elType>::pop() {
   list_element *preservedTop;
   elType toBeReturned;
   if (top!=NULL) {
      // The list is not empty
      toBeReturned = top->contents;
      length--;
      if (cur==top) {
         // The current element is the one that is about to be removed
         cur = cur->next;
      }
      preservedTop = top;
      top = top->next;
      delete preservedTop;
      if (top==NULL) {
         // If top is null, the list is now empty
         bottom = NULL;
      }
      else {
         top->previous = NULL;
      }
   }
   // Returned value can be non-significant if the list is empty
   return (toBeReturned);
}

// Same as pop
template <class elType> elType List<elType>::dequeue() {
   return (pop());
}

// Returns the length of the list
template <class elType> unsigned long long int List<elType>::len() {
   return length;
}

/* Removes the current element from the list; does nothing if the list is empty;
   current element is moved forward (if possible) */
template <class elType> void List<elType>::removeCurrent() {
   list_element *temp;
   if (cur!=NULL) {
      temp = cur;
      if (temp->previous!=NULL) {
         // This is not the first element
         temp->previous->next = temp->next;
      }
      else {
         // This is the first element
         top = temp->next;
      }
      if (temp->next!=NULL) {
         // This is not the last element
         temp->next->previous = temp->previous;
         cur = temp->next;
      }
      else {
         // This is the last element
         bottom = temp->previous;
         cur = temp->previous;
      }
      length--;
      delete temp;
   }
}

/* Sorts list elements, using the merge sort algorithm; sorting
   order is ascending (a reversed sort can easily be obtained
   by exploring the list in reversed order); ordering exploits
   '>' operator, so redefine it in order to obtain a different
   arrangement */
template <class elType> void List<elType>::sort() {
   List< List<elType>* > overallList;
   List<elType> *tempList, *l1, *l2;
   list_element *current;
   bool mergeLists, contl1, contl2;

   if (length>1ULL) {
      // Initial step: build many lists of length 2 from the original data
      current = top;
      while (current!=NULL) {
         tempList = new List<elType>;
         tempList->enqueue(current->contents);
         current = current->next;
         if (current!=NULL) {
            // Compare with the following element
            if (current->previous->contents > current->contents) {
               tempList->push(current->contents);
            }
            else {
               tempList->enqueue(current->contents);
            }
            overallList.enqueue(tempList);
            current = current->next;
         }
      }

      // Now go on by merging the contents of all the lists
      while (overallList.len()>=2) {
         mergeLists = true;
         overallList.goToFirst();
         while (mergeLists) {
            l1 = overallList.current();
            if (mergeLists = overallList.next()) {
               /* There still is a list to merge with; merging will proceed in
                  the following way: overallLists' elements are taken 2 by 2 and
                  merged together; the result is stored in the first of the two
                  elements; the second one is deleted */
               l2 = overallList.current();
               l1->goToFirst();
               l2->goToFirst();
               contl1 = contl2 = true;
               while (contl1 && contl2) {
                  if (l1->current() > l2->current()) {
                     l1->insertB(l2->current());
                     contl2 = l2->next();
                  }
                  else {
                     contl1 = l1->next();
                  }
               }
               while (contl2) {
                  l1->enqueue(l2->current());
                  contl2 = l2->next();
               }
               // Merge is now complete: delete second element
               delete l2;
               overallList.removeCurrent();
            }
         }
      }

      // Copy back the result (stored in the first element of overallList)
      (*this) = *(overallList.first());
      delete overallList.first();
   }
}

// ***************  List  ***************


#endif
