/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: parser.h                                     */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#ifndef PARSER_H
#define PARSER_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

// ******************************************
// ************** File parser ***************
// ******************************************

enum parse_stat {PARSE_NOERROR = 0,             // Parsing was successful
                 PARSE_UNEXPECTED_EOF,          /* File ended unexpectedly (e.g.
                                                   before a `semantic block' could
                                                   be successfully parsed */
                 PARSE_NOLINE,                  /* There is no current line to
                                                   be parsed; this also happens
                                                   when a Parser object has
                                                   just been created, and someone
                                                   attempted to parse its current
                                                   line without previously
                                                   invoking nextLine */
                 PARSE_ILLEGAL_VALUE,           /* Any of the fields of the
                                                   current line has an illegal
                                                   value (e.g. an Autonomous
                                                   System number is 70000) */
                 PARSE_MISMATCH,                /* Some mandatory field is missing
                                                   in the current line */
                 PARSE_NOFILE};                 // No file is currently open

enum parsefile_stat {PARSEFILE_NOERROR = 0,        // Movement to next line was successful
                     PARSEFILE_EOF,                /* There are no more lines to be
                                                      parsed; this is a normal exit
                                                      condition after a file ends */
                     PARSEFILE_NOFILE};            // No file is currently open

class Parser {

   protected:
      unsigned long int parsedlines, readlines; /* readLines stores the number of lines got from
                                                   the file; parsedLines stores how many of them
                                                   have actually been parsed */
      unsigned long int parseFileSize,          // Stores file size (in bytes)
                        parseFilePos;           // Stores current position in the file (in bytes)
      FILE *parseFile;
      bool nowParsing;                          /* This is a switch that toggles parsing
                                                   according to the position in the file
                                                   (e.g., it is used to skip comment areas) */
      int fieldPos[6];                          /* This array stores the column in which each
                                                   field is located */
      char curLine[513];                        // This stores last read line (513 includes `\0')
      char copiedLine[513];                     /* This stores a copy of the last read line,
                                                   that is either returned to user or
                                                   used for internal processing */
      char error[513];                          /* This variable stores a short explanation
                                                   of error conditions (when they occur) */

   public:
      struct parsed_data;                       /* This structure holds data obtained from
                                                   last parsed line */
      Parser();                                 // Default constructor (common to all parsers)
      virtual ~Parser();                        // Destructor
      bool openFile(char*);                     /* Start parsing a given file; returns true if
                                                   file access was successful; false otherwise
                                                   (this also happens when trying to open a
                                                   file while there an already open one);
                                                   this method is common to all parsers */
      void openStdin();                         /* Assume input to be parsed comes from stdin;
                                                   this method is common to all parsers */
      virtual parse_stat parseLine() = 0;       /* Attempts to parse current line; before doing
                                                   that, you must use nextLine in order to move
                                                   over the first parsable line; if you forget
                                                   doing so, a PARSE_NOLINE error will be returned;
                                                   for a full list of possible returned values, see
                                                   above */
      virtual parsefile_stat nextLine() = 0;    /* Moves to the next parsable line in the file;
                                                   for a list of the possible returned values
                                                   see above */
      void closeFile();                         /* Stop parsing and close opened file (do not use it
                                                   if working on stdin); this also resets counters
                                                   (parsed lines, etc.); this method is common
                                                   to all parsers */
      unsigned long int parsedLines();          /* Returns the number of lines that have been
                                                   successfully parsed (e.g. contain useful data)
                                                   until now; this method is common to all parsers */
      unsigned long int readLines();            /* Returns the number of total lines that have
                                                   been read from the file until now (including
                                                   unuseful ones); this method is common to all
                                                   parsers */
      char *errReason();                        /* If parseLine failed, this method returns a
                                                   short explanation of the reason of the fault */

      char *Parser::currentLine();              /* This method returns the line the parser
                                                   is currently on (if any) */
      unsigned long int filePos();              // Returns current position in the file, in bytes
      unsigned long int fileSize();             // Returns the size of the file being parsed, in bytes
};

// ************** File parser ***************

#endif
