/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: policygraph.cc                               */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#include "policygraph.h"

// *********************************************
// ************** BGP Policy graph *************
// *********************************************

/* Warning: using this code assumes that your compiler supports partial
   expression evaluation (e.g., in (A || B), if A is true, B is not
   evaluated at all) */

/* A simple support method that releases memory (if it had been
   allocated before) */
void Policy::release_mem(char **mem_zone) {
   if (*mem_zone != NULL) {
      delete [] *mem_zone;
      *mem_zone = NULL;
   }
}

/* This structure is only used to support automatic memory
   allocation/release for edge labels */

// Default constructor
Policy::Policy() {
   import_policy = new char;
   import_policy[0] = '\0';      // Setup an empty policy
   export_policy = new char;
   export_policy[0] = '\0';      // Setup an empty policy
   oriented = false;
}

// Copy constructor
Policy::Policy(const Policy &otherPolicy) {
   oriented = otherPolicy.oriented;
   import_policy = new char[strlen(otherPolicy.import_policy) + 1]; // One more character for '\0'
   strcpy (import_policy, otherPolicy.import_policy);
   export_policy = new char[strlen(otherPolicy.export_policy) + 1]; // One more character for '\0'
   strcpy (export_policy, otherPolicy.export_policy);
}

// Destructor
Policy::~Policy() {
   Policy::release_mem (&import_policy);
   Policy::release_mem (&export_policy);
}

// Assignment operator
Policy& Policy::operator=(const Policy &otherPolicy) {
   release_mem (&import_policy);
   release_mem (&export_policy);
   import_policy = new char[strlen(otherPolicy.import_policy) + 1]; // One more character for \0
   strcpy (import_policy, otherPolicy.import_policy);
   export_policy = new char[strlen(otherPolicy.export_policy) + 1]; // One more character for \0
   strcpy (export_policy, otherPolicy.export_policy);
   oriented = otherPolicy.oriented;
   return (*this);
}

// Code for the PolicyGraph class begins here

/* The following constant stores the 3-characters tag that identifies
   policy graphs when being written to file */
const char PolicyGraph::policyGraphMarker[4] = "TPG";

/* This method can be used to set the import/export policy of an
   AS. Considering an edge AS1-AS2, this label represents
   the import/export policy of AS1 */
void Policy::setPolicy(policy_type pol, char* policy) {
   switch (pol) {
      case POL_IMPORT:
         release_mem (&import_policy);
         import_policy = new char[strlen(policy) + 1]; // One more character for '\0'
         strcpy (import_policy, policy);
         break;
      case POL_EXPORT:
         release_mem (&export_policy);
         export_policy = new char[strlen(policy) + 1]; // One more character for '\0'
         strcpy (export_policy, policy);
         break;
      default:
         // User must have made some mistake: forgive him by doing nothing
         break;
   }
}

/* This method returns the current import/export policy of an AS,
   if any (returns an empty string if a policy has not been set yet).
   Given an edge AS1-AS2, the method returns the import/export
   policy of AS1. */
char* Policy::getPolicy(policy_type pol) {
   switch (pol) {
      case POL_IMPORT:
         return import_policy;
         break;
      case POL_EXPORT:
         return export_policy;
         break;
      default:
         // Absent-minded user: policy type is invalid, hence we do nothing
         break;
   }
   // Dummy return (just for suppressing compiler warnings)
   return NULL;
}






// Default constructor
PolicyGraph::PolicyGraph() {
   enhanced = false;
   strcpy (error, "");
}

// Destructor
PolicyGraph::~PolicyGraph() {
}


/* This funcion converts an edge direction (IN, OUT, NONE)
   into a printable integer value (-1, 1, 0) */
int PolicyGraph::dirToVal(edge_dir dir) {
   switch (dir) {
      case DIR_IN:   return -1;
                     break;
      case DIR_OUT:  return 1;
                     break;
      case DIR_NONE: return 0;
                     break;
      default:       return 0;
                     break;
   }
}

/* This function converts an integer value (-1, 1, 0)
   into an edge direction (IN, OUT, NONE) */
edge_dir PolicyGraph::valToDir(int val) {
   switch(val) {
      case -1: return DIR_IN;
               break;
      case 1:  return DIR_OUT;
               break;
      case 0:  return DIR_NONE;
               break;
      default: return DIR_NONE;
               break;
   }
}



/* Dumps all the policy graph contents to an
   ASCII file; use `readFromFile' to
   get them back; the only parameter is
   a pointer to an open file strem;
   this method returns true if operation
   is successful; false otherwise */
bool PolicyGraph::writeToFile(FILE *f) {
   List <graph_node*> nodeList;
   List <graph_edge*> edgeList;
   graph_node *curNode;
   graph_edge *curEdge;
   bool stop, goOn;
   
   strcpy (error, "");
   // Write the string that identifies the graph
   if (!fprintf (f, "%s\n", policyGraphMarker)) {
      strcat (error, "Could not write file type string");
      return false;
   }
   
   // Write number of nodes
   if (!fprintf (f, "%u\n", nodes())) {
      strcat (error, "Could not write number of nodes");
      return false;
   }

   // Write number of edges
   if (!fprintf (f, "%u\n", edges())) {
      strcat (error, "Could not write number of edges");
      return false;
   }

   // Write list of nodes
   nodeList = listNodes();
   stop = false;
   goOn = !nodeList.isEmpty();
   for (nodeList.goToFirst(); goOn; goOn = nodeList.next()) {
      curNode = nodeList.current();
      if (!stop && !fprintf (f, "%i\n", *(curNode->contents))) {
         strcat (error, "Could not write node");
         sprintf(error, "%s %i", error, *(curNode->contents));
         /* There is no immediate return, because that would not
            allow memory deallocation */
         stop = true;
      }
      delete curNode;
   }
   
   if (stop) return false;
   
   // Write list of edges
   edgeList = listEdges();
   goOn = !edgeList.isEmpty();
   for (edgeList.goToFirst(); goOn; goOn = edgeList.next()) {
      curEdge = edgeList.current();
      if (!stop && !fprintf (f, "%i %i %i %i %i %s %i %s %i %s %i %s\n",
            *(curEdge->from->contents),
            *(curEdge->to->contents),
            (curEdge->weight_from->oriented ? 1 : 0),
            dirToVal(*(curEdge->orientation)),
            strlen(curEdge->weight_from->getPolicy(POL_IMPORT)),
            curEdge->weight_from->getPolicy(POL_IMPORT),
            strlen(curEdge->weight_from->getPolicy(POL_EXPORT)),
            curEdge->weight_from->getPolicy(POL_EXPORT),
            strlen(curEdge->weight_to->getPolicy(POL_IMPORT)),
            curEdge->weight_to->getPolicy(POL_IMPORT),
            strlen(curEdge->weight_to->getPolicy(POL_EXPORT)),
            curEdge->weight_to->getPolicy(POL_EXPORT))) {
         strcat (error, "Could not write edge");
         sprintf(error, "%s %i %i", error, *(curEdge->from->contents),
                                          *(curEdge->to->contents));
         /* There is no immediate return, because that would not
            allow memory deallocation */
         stop = true;
      }

      delete curEdge->from;
      delete curEdge->to;
      delete curEdge;
   }

   return true;
}

/* Builds a policy graph from the data that have
   previously been generated by `writeToFile';
   the only parameter is a pointer to an
   open file stream; returns true if
   operation is successful; false otherwise */
bool PolicyGraph::readFromFile(FILE *f) {
   char string[513], oldString[513];             // One more byte for '\0'
   char *err, *curToken;
   int intValue;
   unsigned int uValue, nodeCnt, edgeCnt;
   unsigned long int ulValue, policyLength;
   ASnode from, to;
   Policy pol1, pol2;
   edge_dir direction;


   strcpy (error, "");

   // Get graph type
   if (!fgets(string, 513, f)) {
      strcat (error, "Could not retrieve file type string");
      return false;
   }
   
   // Remove CR/LF at end of line
   if (strpbrk(string, "\n\r")) {
      *(strpbrk(string, "\n\r")) = '\0';
   }
   if (strcmp(string, policyGraphMarker) != 0) {
      strcat (error, "Not a policy graph");
      return false;
   }

      // Get number of nodes
   if (!fgets(string, 512, f)) {
      strcat (error, "Could not retrieve number of nodes");
      return false;
   }
   // Remove CR/LF at end of line
   if (strpbrk(string, "\n\r")) {
      *(strpbrk(string, "\n\r")) = '\0';
   }
   uValue = strtoul(string, &err, 10);
   if (strlen(err) > 0) {
      strcat (error, "Invalid number of nodes: `");
      strcat (error, string);
      strcat (error, "'");
      return false;
   }
   nodeCnt = uValue;
   
   // Get number of edges
   if (!fgets(string, 512, f)) {
      strcat (error, "Could not retrieve number of edges");
      return false;
   }
   // Remove CR/LF at end of line
   if (strpbrk(string, "\n\r")) {
      *(strpbrk(string, "\n\r")) = '\0';
   }
   uValue = strtoul(string, &err, 10);
   if (strlen(err) > 0) {
      strcat (error, "Invalid number of edges: `");
      strcat (error, string);
      strcat (error, "'");
      return false;
   }
   edgeCnt = uValue;

   // Get nodes
   for (unsigned int i = 0; nodeCnt > 0 && i <= nodeCnt - 1; i++) {
      if (!fgets(string, 512, f)) {
         strcat (error, "Unable to retrieve all the nodes");
         return false;
      }
      // Remove CR/LF at end of line
      if (strpbrk(string, "\n\r")) {
         *(strpbrk(string, "\n\r")) = '\0';
      }
      intValue = strtol(string, &err, 10);
      if (intValue < 1 || intValue > 65535 || (strlen(err) > 0)) {
         strcat (error, "Invalid AS number: `");
         strcat (error, string);
         strcat (error, "'");
         return false;
      }
      addNode((ASnode) intValue, (ASnode) intValue);
   }
   
   // Get edges
   for (unsigned int i = 0; edgeCnt > 0 && i <= edgeCnt - 1; i++) {
      if (!fgets(string, 512, f)) {
         strcat (error, "Unable to retrieve all the edges");
         return false;
      }
      // Remove CR/LF at end of line
      if (strpbrk(string, "\n\r")) {
         *(strpbrk(string, "\n\r")) = '\0';
      }
      /* String will be tokenized, so a copy
         of it should be made */
      strcpy(oldString, string);

      // First node
      curToken = strtok(string, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      intValue = strtol(curToken, &err, 10);
      if (intValue < 1 || intValue > 65535 || (strlen(err) > 0)) {
         strcat (error, "Invalid AS number: `");
         strcat (error, curToken);
         strcat (error, "'");
         return false;
      }
      from = intValue;

      // Second node
      curToken = strtok(NULL, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      intValue = strtol(curToken, &err, 10);
      if (intValue < 1 || intValue > 65535 || (strlen(err) > 0)) {
         strcat (error, "Invalid AS number: `");
         strcat (error, curToken);
         strcat (error, "'");
         return false;
      }
      to = intValue;

      // Edge direction flag
      curToken = strtok(NULL, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      if (strcmp(curToken, "0") != 0 && strcmp(curToken, "1") != 0) {
         strcat (error, "Invalid edge direction flag: `");
         strcat (error, curToken);
         strcat (error, "'");
         return false;
      }
      pol1.oriented = pol2.oriented =
         (strcmp(curToken, "1") == 0);

      // Edge direction
      curToken = strtok(NULL, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      intValue = strtol(curToken, &err, 10);
      if (intValue < -1 || intValue > 1 || (strlen(err) > 0)) {
         strcat (error, "Invalid edge direction: `");
         strcat (error, curToken);
         strcat (error, "'");
         return false;
      }
      direction = valToDir(intValue);

      /* The following comments assume these conventions:
         IMPORT-POLICY-FROM: import policy from "this" side of the edge
                             (i.e., given an edge AS1-AS2, this is the
                             import policy of AS1)
         EXPORT-POLICY-FROM: export policy from "this" side of the edge
                             (i.e., given an edge AS1-AS2, this is the
                             export policy of AS1)
         IMPORT-POLICY-TO:   import policy from "the other" side of the edge
                             (i.e., given an edge AS1-AS2, this is the
                             import policy of AS2)
         EXPORT-POLICY-TO:   export policy from "the other" side of the edge
                             (i.e., given an edge AS1-AS2, this is the
                             export policy of AS2) */
                             
      // Length of IMPORT-POLICY-FROM
      curToken = strtok(NULL, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      ulValue = strtoul(curToken, &err, 10);
      if (strlen(err) > 0) {
         sprintf (error, "Invalid import policy length: `%s' for AS %i from edge %i %i",
                  curToken, from, from, to);
         return false;
      }
      policyLength = ulValue;
      
      // IMPORT-POLICY-FROM
      curToken += (strlen(curToken) + 1);
      if (curToken > string + strlen(oldString)) {
         sprintf (error, "Could not get import policy for AS %i from edge %i %i",
                  from, from, to);
         return false;
      }
      curToken[policyLength] = '\0';      // Split string at the end of the current policy
      pol1.setPolicy(POL_IMPORT, curToken);
      
      // Length of EXPORT-POLICY-FROM
      curToken += (strlen(curToken) + 1);
      curToken = strtok(curToken, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      ulValue = strtoul(curToken, &err, 10);
      if (strlen(err) > 0) {
         sprintf (error, "Invalid export policy length: `%s' for AS %i from edge %i %i",
                  curToken, from, from, to);
         return false;
      }
      policyLength = ulValue;
      
      // EXPORT-POLICY-FROM
      curToken += (strlen(curToken) + 1);
      if (curToken > string + strlen(oldString)) {
         sprintf (error, "Could not get export policy for AS %i from edge %i %i",
                  from, from, to);
         return false;
      }
      curToken[policyLength] = '\0';      // Split string at the end of the current policy
      pol1.setPolicy(POL_EXPORT, curToken);

      // Length of IMPORT-POLICY-TO
      curToken += (strlen(curToken) + 1);
      curToken = strtok(curToken, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      ulValue = strtoul(curToken, &err, 10);
      if (strlen(err) > 0) {
         sprintf (error, "Invalid import policy length: `%s' for AS %i from edge %i %i",
                  curToken, to, from, to);
         return false;
      }
      policyLength = ulValue;
      
      // IMPORT-POLICY-TO
      curToken += (strlen(curToken) + 1);
      if (curToken > string + strlen(oldString)) {
         sprintf (error, "Could not get import policy for AS %i from edge %i %i",
                  to, from, to);
         return false;
      }
      curToken[policyLength] = '\0';      // Split string at the end of the current policy
      pol2.setPolicy(POL_IMPORT, curToken);

      // Length of EXPORT-POLICY-TO
      curToken += (strlen(curToken) + 1);
      curToken = strtok(curToken, " ");
      if (curToken == NULL) {
         strcat (error, "Invalid edge line format: `");
         strcat (error, oldString);
         strcat (error, "'");
         return false;
      }
      ulValue = strtoul(curToken, &err, 10);
      if (strlen(err) > 0) {
         sprintf (error, "Invalid export policy length: `%s' for AS %i from edge %i %i",
                  curToken, to, from, to);
         return false;
      }
      policyLength = ulValue;
      
      // EXPORT-POLICY-TO
      curToken += (strlen(curToken) + 1);
      if (curToken > string + strlen(oldString)) {
         sprintf (error, "Could not get export policy for AS %i from edge %i %i",
                  to, from, to);
         return false;
      }
      curToken[policyLength] = '\0';      // Split string at the end of the current policy
      pol2.setPolicy(POL_EXPORT, curToken);
      
      addEdge(from, to, direction, pol1, pol2);
      
   //////////////////////////////////////////////
//      printf ("Read edge: %i %i '%s' '%s' '%s' '%s'\n", from, to, pol1.getPolicy(POL_IMPORT), pol1.getPolicy(POL_EXPORT),
//              pol2.getPolicy(POL_IMPORT), pol2.getPolicy(POL_EXPORT));
   }
   
   return true;
}

/* Returns a string that describes the
   reason of last file error; this method
   should be invoked after that either one
   of `writeToFile' or `readFromFile'
   returned false */
char* PolicyGraph::errReason() {
   return error;
}

// ************** BGP Policy graph *************
