/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: smartstructs.h                               */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#ifndef SMARTSTRUCTS_H
#define SMARTSTRUCTS_H

#include "list.h"

// *************************************************
// ******** Binary Search Tree (AVL tree) **********
// *************************************************

template <class elType, class keyType> class BST {
   private:
      struct tree_node {
         elType contents;                    // Node contents
         keyType key;                        // Element key; item ordering is based upon it
         tree_node *left, *right;            // Node successors
         char unbalancing;                   /* This variable stores the difference between
                                                the depth of the right subtree and the one
                                                of the left subtree; it is used in order
                                                to always keep the tree balanced */
         unsigned char visited;              /* This is a bit mask that is used during tree
                                                traversal */
      };
      tree_node *root;                       // Tree root (base node)
      unsigned long long int nodeCount;      // Traces number of nodes
      void rotateLL (tree_node *current,
                     tree_node *parent,
                     bool isLeftChild);      /* Perform a LL rotation; 'current' is the node
                                                to be rotated; 'parent' is its parent node;
                                                'isLeftChild' must be true if 'current' is
                                                'parent''s left child; false otherwise */
      void rotateRR (tree_node *current,
                     tree_node *parent,
                     bool isLeftChild);      /* Perform a RR rotation; parameters are the same
                                                as for rotateLL */
      void rotateLR (tree_node *current,
                     tree_node *parent,
                     bool isLeftChild);      /* Perform a LR rotation; parameters are the same
                                                as for rotateLL */
      void rotateRL (tree_node *current,
                     tree_node *parent,
                     bool isLeftChild);      /* Perform a RL rotation; parameters are the same
                                                as for rotateLL */
      void deleteSubTree(tree_node*);        /* Releases memory for the whole subtree starting
                                                at a specified node */
      List<elType> retItemList;              // These variables are used when returning a list
      List<keyType> retKeyList;              // of the nodes/keys in the tree
   public:
      /* Ordering of the elements inside the tree is based exclusively upon
         the '>' and '==' operators, so redefine them in order to obtain a
         different arrangement */
      BST();                                 // Default constructor
      ~BST();                                // Destructor
      BST(const BST&);                       // Copy constructor
      BST& operator=(const BST&);            // Assignment operator overload
      bool isEmpty();                        // Returns true if the tree contains no nodes
      unsigned long long int items();        // Returns the number of nodes inside the tree
      bool addItem(const elType&,
                   const keyType&);          /* Insert a new node inside the tree
                                                (its insertion point is determined by the
                                                user supplied key); does nothing (and
                                                returns false) if a node with the same key
                                                is already in the tree; in all the other
                                                cases it returns true */
      bool deleteItem(const keyType&);       /* Removes a node from the tree, searching
                                                it by using its key value; returns true
                                                if deletion was successful; false
                                                otherwise (node could not be found) */
      List<elType>& listItems();             /* List tree contents (ordering is coherent
                                                with key values) */
      List<keyType>& listKeys();             // List keys in the tree (in an ordered fashion)
      elType* findItem(const keyType&);      /* Checks whether a node exists; if so, it
                                                returns a pointer to its contents;
                                                otherwise, it returns a NULL pointer */
};





/* Ordering of the elements inside the tree is based exclusively upon
   the '>' and '==' operators, so redefine them in order to obtain a
   different arrangement */

/* Perform a LL rotation; 'current' is the node to be rotated; 'parent' is
   its parent node; 'isLeftChild' must be true if 'current' is 'parent''s
   left child; false otherwise */
template <class elType, class keyType>
   void BST<elType, keyType>::rotateLL (tree_node *current, tree_node *parent,
                                        bool isLeftChild) {
   tree_node *oldSubTree;
   oldSubTree = current->left;
   current->left = current->left->right;
   if (parent==NULL) {
      root = oldSubTree;
   }
   else {
      if (isLeftChild) {
         parent->left = oldSubTree;
      }
      else {
         parent->right = oldSubTree;
      }
   }
   oldSubTree->right = current;
   // Update unbalancing fields
   current->unbalancing = -1 - oldSubTree->unbalancing;
   oldSubTree->unbalancing++;
}

// Perform a RR rotation; parameters are the same as for rotateLL
template <class elType, class keyType>
   void BST<elType, keyType>::rotateRR (tree_node *current, tree_node *parent,
                                        bool isLeftChild) {
   tree_node *oldSubTree;
   oldSubTree = current->right;
   current->right = current->right->left;
   if (parent==NULL) {
      root = oldSubTree;
   }
   else {
      if (isLeftChild) {
         parent->left = oldSubTree;
      }
      else {
         parent->right = oldSubTree;
      }
   }
   oldSubTree->left = current;
   // Update unbalancing fields
   current->unbalancing = 1 - oldSubTree->unbalancing;
   oldSubTree->unbalancing--;
}

// Perform a LR rotation; parameters are the same as for rotateLL
template <class elType, class keyType>
   void BST<elType, keyType>::rotateLR (tree_node *current, tree_node *parent,
                                        bool isLeftChild) {
   tree_node *oldLeftSubTree, *oldRightSubTree;
   oldLeftSubTree = current->left;
   current->left = current->left->right->right;
   oldRightSubTree = oldLeftSubTree->right;
   oldLeftSubTree->right = oldLeftSubTree->right->left;
   oldRightSubTree->left = oldLeftSubTree;
   oldRightSubTree->right = current;
   if (parent==NULL) {
      root = oldRightSubTree;
   }
   else {
      if (isLeftChild) {
         parent->left = oldRightSubTree;
      }
      else {
         parent->right = oldRightSubTree;
      }
   }
   // Update unbalancing fields
   current->unbalancing =
      -(oldRightSubTree->unbalancing < 0 ? oldRightSubTree->unbalancing : 0);
   oldLeftSubTree->unbalancing =
      -(oldRightSubTree->unbalancing > 0 ? oldRightSubTree->unbalancing : 0);
   oldRightSubTree->unbalancing = 0;
}

// Perform a RL rotation; parameters are the same as for rotateLL
template <class elType, class keyType>
   void BST<elType, keyType>::rotateRL (tree_node *current, tree_node *parent,
                                        bool isLeftChild) {
   tree_node *oldLeftSubTree, *oldRightSubTree;
   oldRightSubTree = current->right;
   current->right = current->right->left->left;
   oldLeftSubTree = oldRightSubTree->left;
   oldRightSubTree->left = oldRightSubTree->left->right;
   oldLeftSubTree->right = oldRightSubTree;
   oldLeftSubTree->left = current;
   if (parent==NULL) {
      root = oldLeftSubTree;
   }
   else {
      if (isLeftChild) {
         parent->left = oldLeftSubTree;
      }
      else {
         parent->right = oldLeftSubTree;
      }
   }
   // Update unbalancing fields
   current->unbalancing =
      -(oldLeftSubTree->unbalancing > 0 ? oldLeftSubTree->unbalancing : 0);
   oldRightSubTree->unbalancing =
      -(oldLeftSubTree->unbalancing < 0 ? oldLeftSubTree->unbalancing : 0);
   oldLeftSubTree->unbalancing = 0;
}

// Releases memory for the whole subtree starting at a specified node
template <class elType, class keyType>
   void BST<elType, keyType>::deleteSubTree(tree_node *start) {
   List<tree_node*> toBeVisited;
   tree_node *current;
   if (start!=NULL) {
      toBeVisited.enqueue(start);
      while (!toBeVisited.isEmpty()) {
         current = toBeVisited.dequeue();
         // Avoid trying to visit nodes that do not exist
         if (current->left!=NULL) {
            toBeVisited.enqueue(current->left);
         }
         if (current->right!=NULL) {
            toBeVisited.enqueue(current->right);
         }
         delete current;
      }
   }
}

// Default constructor
template <class elType, class keyType> BST<elType, keyType>::BST() {
   root = NULL;
   nodeCount = 0ULL;
}

// Destructor
template <class elType, class keyType> BST<elType, keyType>::~BST() {
   deleteSubTree(root);
   /* No need to set root pointer to NULL, because this object does not
      exist any longer */
}

// Copy constructor
template <class elType, class keyType>
   BST<elType, keyType>::BST(const BST& source) {
   List<tree_node*> toBeVisitedSource, toBeVisitedDest;
   tree_node *sourceCurrent, *destCurrent;
   root = NULL;
   if (source.root!=NULL) {
      root = new tree_node;
      toBeVisitedSource.enqueue(source.root);
      toBeVisitedDest.enqueue(root);
      /* Both lists should have the same length throughout the whole
         process, so it is the same to test for toBeVisitedSource.isEmpty()
         or toBeVisitedDest.isEmpty() to be false */
      while (!toBeVisitedSource.isEmpty()) {
         sourceCurrent = toBeVisitedSource.dequeue();
         destCurrent = toBeVisitedDest.dequeue();
         // Copy node data
         destCurrent->contents = sourceCurrent->contents;
         destCurrent->key = sourceCurrent->key;
         destCurrent->unbalancing = sourceCurrent->unbalancing;
         destCurrent->left = NULL;
         destCurrent->right = NULL;
         // Explore left subtree
         if (sourceCurrent->left!=NULL) {
            destCurrent->left = new tree_node;
            toBeVisitedSource.enqueue(sourceCurrent->left);
            toBeVisitedDest.enqueue(destCurrent->left);
         }
         // Explore right subtree
         if (sourceCurrent->right!=NULL) {
            destCurrent->right = new tree_node;
            toBeVisitedSource.enqueue(sourceCurrent->right);
            toBeVisitedDest.enqueue(destCurrent->right);
         }
      }
   }
   nodeCount = source.nodeCount;
}

// Assignment operator overload
template <class elType, class keyType>
   BST<elType, keyType>& BST<elType, keyType>::operator=(const BST& source) {
   List<tree_node*> toBeVisitedSource, toBeVisitedDest;
   tree_node *sourceCurrent, *destCurrent;
   if (source.root==NULL) {
      // Source tree is empty: delete current tree
      deleteSubTree(root);
      root = NULL;
   }
   else {
      // Source tree is not empty: if necessary, create a new root node
      if (root==NULL) {
         root = new tree_node;
         /* Links to successor nodes cannot be overwritten by default,
            since the structure of the old tree is not to be deleted, but
            shrunk or expanded in order to contain the new data */
         root->left = NULL;
         root->right = NULL;
      }
      toBeVisitedSource.enqueue(source.root);
      toBeVisitedDest.enqueue(root);
      while (!toBeVisitedSource.isEmpty()) {
         sourceCurrent = toBeVisitedSource.dequeue();
         destCurrent = toBeVisitedDest.dequeue();
         // Copy node data
         destCurrent->contents = sourceCurrent->contents;
         destCurrent->key = sourceCurrent->key;
         destCurrent->unbalancing = sourceCurrent->unbalancing;
         // Explore left subtree
         if (sourceCurrent->left!=NULL) {
            toBeVisitedSource.enqueue(sourceCurrent->left);
            if (destCurrent->left==NULL) {
               // The node did not exist in the destination tree: create it
               destCurrent->left = new tree_node;
               destCurrent->left->left = NULL;
               destCurrent->left->right = NULL;
            }
            toBeVisitedDest.enqueue(destCurrent->left);
         }
         else {
            /* Left subtree does not exist in the source tree: delete it
               also from the current tree */
            deleteSubTree(destCurrent->left);
            destCurrent->left = NULL;
         }
         // Explore right subtree
         if (sourceCurrent->right!=NULL) {
            toBeVisitedSource.enqueue(sourceCurrent->right);
            if (destCurrent->right==NULL) {
               // The node did not exist in the destination tree: create it
               destCurrent->right = new tree_node;
               destCurrent->right->left = NULL;
               destCurrent->right->right = NULL;
            }
            toBeVisitedDest.enqueue(destCurrent->right);
         }
         else {
            /* Right subtree does not exist in the source tree: delete it
               also from the current tree */
            deleteSubTree(destCurrent->right);
            destCurrent->right = NULL;
         }
      }
   }
   nodeCount = source.nodeCount;
   return (*this);
}

// Returns true if the tree contains no nodes
template <class elType, class keyType> bool BST<elType, keyType>::isEmpty() {
   return (root==NULL);
}

// Returns the number of nodes inside the tree
template <class elType, class keyType>
   unsigned long long int BST<elType, keyType>::items() {
   return (nodeCount);
}

/* Insert a new node inside the tree (its insertion point
   is determined by the user supplied key); does nothing
   (and returns false) if a node with the same key is
   already in the tree; in all the other cases it returns true */
template <class elType, class keyType>
   bool BST<elType, keyType>::addItem(const elType& item, const keyType& key) {
   List<tree_node*> toBeUpdated;
   tree_node *current, *parent;
   bool isLeftChild, increasedHeight;
   if (root==NULL) {
      // Create a new tree
      root = new tree_node;
      root->contents = item;
      root->key = key;
      root->unbalancing = 0;
      root->left = NULL;
      root->right = NULL;
      nodeCount++;
   }
   else {
      // Insert a node into an existing tree
      current = root;
      while (current!=NULL) {
         if (current->key == key)  {
            // Trying to insert a duplicate node
            return false;
         }
         /* Keep trace of visited nodes in order to update their unbalancing
            field later on */
         toBeUpdated.push(current);
         if (current->key > key) {
            // Continue in the left subtree or create a new one
            if (current->left==NULL) {
               // Create a new node (that is the left subtree of the current node)
               current->left = new tree_node;
               current->left->contents = item;
               current->left->key = key;
               current->left->left = NULL;
               current->left->right = NULL;
               current->left->unbalancing = 0;
               // Terminate cycle
               current = NULL;
            }
            else {
               current = current->left;
            }
         }
         else {
            // Continue in the right subtree or create a new one
            if (current->right==NULL) {
               // Create a new node (that is the right subtree of the current node)
               current->right = new tree_node;
               current->right->contents = item;
               current->right->key = key;
               current->right->left = NULL;
               current->right->right = NULL;
               current->right->unbalancing = 0;
               // Terminate cycle
               current = NULL;
            }
            else {
               current = current->right;
            }
         }
      }
      nodeCount++;

      /* A new node has been inserted: update unbalancing fields and
         rebalance tree */
      increasedHeight = true;
      while (!toBeUpdated.isEmpty()) {
         current = toBeUpdated.pop();
         // toBeUpdated's top (if exists) contains parent node's address
         if (!toBeUpdated.isEmpty()) {
            parent = toBeUpdated.first();
            isLeftChild = (parent->key > current->key);
         }
         else {
            parent = NULL;
            /* If current node has no parent (root node), isLeftChild's
               value is senseless */
         }
         // Recompute unbalancing field
         if (current->key > key) {
            // Insertion took place in the left subtree
            if (increasedHeight) {
               if (current->unbalancing==1) {
                  increasedHeight = false;      /* Parent nodes don't change their
                                                   unbalancing field */
               }
               current->unbalancing--;
            }
         }
         else {
            // Insertion took place in the right subtree
            if (increasedHeight) {
               if (current->unbalancing==-1) {
                  increasedHeight = false;   /* Parent nodes don't change their
                                                unbalancing field */
               }
               current->unbalancing++;
            }
         }
         // Now rebalance (if necessary)
         if (current->unbalancing < -1) {
            // Current node needs rebalancing
            if (current->left->unbalancing > 0) {
               // Must perform a LR rotation
               rotateLR (current, parent, isLeftChild);
               // LR rotations always compensate tree height increase, so...
               increasedHeight = false;
            }
            else {
               // Must perform a LL rotation
               if (current->left->unbalancing < 0 ) {
                  // In this case, a LL rotation compensates tree height increase, so...
                  increasedHeight = false;
               }
               rotateLL (current, parent, isLeftChild);
            }
         }
         if (current->unbalancing > 1) {
            // Current node needs rebalancing
            if (current->right->unbalancing < 0) {
               // Must perform a RL rotation
               rotateRL (current, parent, isLeftChild);
               // RL rotations always compensate tree height increase, so...
               increasedHeight = false;
            }
            else {
               // Must perform a RR rotation
               if (current->right->unbalancing > 0 ) {
                  // In this case, a RR rotation compensates tree height increase, so...
                  increasedHeight = false;
               }

               rotateRR (current, parent, isLeftChild);
            }
         }
      }
   }
   return true;
}

/* Removes a node from the tree, searching it by using its key
   value; returns true if deletion was successful; false
   otherwise (node could not be found) */
template <class elType, class keyType>
   bool BST<elType, keyType>::deleteItem(const keyType& key) {
   List<tree_node*> toBeUpdated;
   tree_node *current, *parent, *oldCurrent, *oldParent, *replacedNode;
   bool wentToTheLeft, decreasedHeight, done, exploredRight, listWasEmpty;
   current = root;
   oldCurrent = replacedNode = NULL;
   parent = NULL;
   done = false;
   while (!done && current!=NULL) {
      if (current->key == key) {
         // Node to be removed has been found
         nodeCount--;
         done = true;
         if (current->left == NULL && current->right == NULL) {
            // Case A: node is a leaf: discard node and exit
            delete current;
            if (parent==NULL) {
               // Node was root node
               root = NULL;
            }
            else {
               if (wentToTheLeft) {
                  // Current node is parent's left child
                  parent->left = NULL;
               }
               else {
                  // Current node is parent's right child
                  parent->right = NULL;
               }
            }
         }
         else {
            if (current->left != NULL && current->right == NULL) {
               /* Case B: node only has a left child: replace it with its
                  left child's root */
               if (parent==NULL) {
                  // Node was root node
                  root = current->left;
               }
               else {
                  if (wentToTheLeft) {
                     parent->left = current->left;
                  }
                  else {
                     parent->right = current->left;
                  }
               }
               delete current;
            }
            else {
               if (current->right != NULL && current->left == NULL) {
                  /* Case C: node only has a right child: replace it with its
                     right child's root */
                  if (parent==NULL) {
                     // Node was root node
                     root = current->right;
                  }
                  else {
                     if (wentToTheLeft) {
                        parent->left = current->right;
                     }
                     else {
                        parent->right = current->right;
                     }
                  }
                  delete current;
               }
               else {
                  /* Case D: node has both left and right subtrees.
                     Look for left subtree's rightmost node (maximum of the left
                     subtree) and replace such node with the current one (the
                     one to be deleted) */
                  oldCurrent = current;
                  oldParent = parent;
                  /* Remember where to insert node that will take the place of the
                     deleted one; `listWasEmpty' is used because it is impossible
                     to set a "mark" point over an empty list */
                  listWasEmpty = toBeUpdated.isEmpty();
                  toBeUpdated.goToFirst();
                  // Start exploring left subtree
                  exploredRight = false;
                  parent = current;
                  current = current->left;
                  while (current->right != NULL) {
                     exploredRight = true;
                     toBeUpdated.push(current);
                     parent = current;
                     current = current->right;
                  }
                  /* Take the node that has just been found and put it in the place
                     of the one that is about to be deleted */
                  if (exploredRight) {
                     parent->right = current->left;
                  }
                  else {
                     parent->left = current->left;
                  }
                  if (oldParent==NULL) {
                     // Deleting root node
                     root = current;
                  }
                  else {
                     if (wentToTheLeft) {
                        oldParent->left = current;
                     }
                     else {
                        oldParent->right = current;
                     }
                  }
                  current->right = oldCurrent->right;
                  /* Must be careful when reassigning the following pointer:
                     avoid making it point to the node itself!!! */
                  if (exploredRight) {
                     current->left = oldCurrent->left;
                  }
                  current->unbalancing = oldCurrent->unbalancing;
                  /* Also the current node needs to be updated: insert it into
                     the stack at the saved point */
                  if (listWasEmpty) {
                     toBeUpdated.enqueue(current);
                  }
                  else {
                     toBeUpdated.insertB(current);
                  }
                  replacedNode = current;
                  delete oldCurrent;
               }
            }
         }
      }
      else {
         /* Node has not been found yet: store current node for
            later unbalancing field update or rebalancing */
         toBeUpdated.push(current);
         if (current->key > key) {
            // Continue search in the left subtree
            parent = current;
            current = current->left;
            wentToTheLeft = true;
         }
         else {
            // Continue search in the right subtree
            parent = current;
            current = current->right;
            wentToTheLeft = false;
         }
      }
   }

   if (!done) {
      // Node could not be found (this also happens for empty trees)
      return false;
   }

   // Now update balancing fields and rebalance tree, if necessary
   decreasedHeight = true;
   while (!toBeUpdated.isEmpty()) {
      current = toBeUpdated.pop();
      if (!toBeUpdated.isEmpty()) {
         // toBeUpdated's top always stores parent node's address
         parent = toBeUpdated.first();
         wentToTheLeft = (parent->key > current->key);
      }
      else {
         parent = NULL;
      }

      // Update unbalancing field
      /* Note that original tree traversal order is influenced by the
         old node displacement: if a node replacement took place (case D),
         deletion necessarily applied to left subtree */
      if (current==replacedNode || current->key > key) {
         // Deletion took place somewhere in the left subtree
         if (decreasedHeight) {
            if (current->unbalancing==0) {
               decreasedHeight = false;   /* Nodes above the current one don't
                                             need to have their unbalancing field
                                             updated */
            }
            current->unbalancing++;
         }
      }
      else {
         // Deletion took place somewhere in the right subtree
         if (decreasedHeight) {
            if (current->unbalancing==0) {
               decreasedHeight = false;   /* Nodes above the current one don't
                                             need to have their unbalancing field
                                             updated */
            }
            current->unbalancing--;
         }
      }


      // Now apply rotation, if necessary
      if (current->unbalancing < -1) {
         // Current node needs rebalancing
         if (current->left->unbalancing > 0) {
            // Must perform a LR rotation
            rotateLR (current, parent, wentToTheLeft);
            /* A mixed rotation always decreases tree height by 1, so
               decreasedHeight must not be set to false */
         }
         else {
            // Must perform a LL rotation
            if (current->left->unbalancing == 0) {
               // This rotation won't compensate tree height decrease
               decreasedHeight = false;
            }
            rotateLL (current, parent, wentToTheLeft);
         }
      }
      if (current->unbalancing > 1) {
         // Current node needs rebalancing
         if (current->right->unbalancing < 0) {
            // Must perform a RL rotation
            rotateRL (current, parent, wentToTheLeft);
            /* A mixed rotation always decreases tree height by 1, so
               decreasedHeight must not be set to false */
         }
         else {
            // Must perform a RR rotation
            if (current->right->unbalancing == 0) {
               // This rotation won't compensate tree height decrease
               decreasedHeight = false;
            }
            rotateRR (current, parent, wentToTheLeft);
         }
      }
   }
   return true;
}

// List tree contents (ordering is coherent with key values)
template <class elType, class keyType>
   List<elType>& BST<elType, keyType>::listItems() {
   List<tree_node*> toBeVisited;
   tree_node *current;
   // Initialize an internal empty list
   retItemList = List<elType>();
   if (root!=NULL) {
      // Inorder tree visit (left subtree, then current node, then right subtree)
      /* Exploit 'visited' bits to trace visit order:
         00000000 = Visit left subtree
         00000001 = Visit current node
         00000010 = Visit right subtree */
      root->visited &= 0x0;                        // 00000000 = Visit left subtree
      toBeVisited.push(root);
      while (!toBeVisited.isEmpty()) {
         current = toBeVisited.pop();
         switch(current->visited) {
            case 0x0:   // 00000000 = Visit left subtree
               current->visited += 0x1;            // Next time visit current node
               toBeVisited.push(current);
               // Explore left subtree first
               if (current->left!=NULL) {
                  current->left->visited &= 0x0;   // 00000000 = Visit left subtree
                  toBeVisited.push(current->left);
               }
               break;
            case 0x1: // 00001000 = Visit current node
               current->visited += 0x1;            // Next time visit right subtree
               toBeVisited.push(current);
               retItemList.enqueue(current->contents);
               break;
            case 0x2:   // 00000010 = Visit right subtree
               /* This node must not be visited any more: do not insert
                  it in the list of nodes to be visited */
               current->visited &= 0x0;            /* 00000000 (mask that resets visit
                                                      order bits to 0) */
               if (current->right!=NULL) {
                  current->right->visited &= 0x0;  // 00000000 = Visit left subtree
                  toBeVisited.push(current->right);
               }
               break;
            default:
               break;
         }
      }
   }
   return retItemList;
}

// List keys in the tree (in an ordered fashion)
template <class elType, class keyType>
   List<keyType>& BST<elType, keyType>::listKeys() {
   List<tree_node*> toBeVisited;
   tree_node *current;
   // Initialize an internal empty list
   retKeyList = List<keyType>();
   if (root!=NULL) {
      // Inorder tree visit (left subtree, then current node, then right subtree)
      /* Exploit 'visited' bits to trace visit order:
         00000000 = Visit left subtree
         00000001 = Visit current node
         00000010 = Visit right subtree */
      root->visited &= 0x0;                        // 00000000 = Visit left subtree
      toBeVisited.push(root);
      while (!toBeVisited.isEmpty()) {
         current = toBeVisited.pop();
         switch(current->visited) {
            case 0x0:   // 00000000 = Visit left subtree
               current->visited += 0x1;            // Next time visit current node
               toBeVisited.push(current);
               // Explore left subtree first
               if (current->left!=NULL) {
                  current->left->visited &= 0x0;   // 00000000 = Visit left subtree
                  toBeVisited.push(current->left);
               }
               break;
            case 0x1: // 00001000 = Visit current node
               current->visited += 0x1;            // Next time visit right subtree
               toBeVisited.push(current);
               retKeyList.enqueue(current->key);
               break;
            case 0x2:   // 00000010 = Visit right subtree
               /* This node must not be visited any more: do not insert
                  it in the list of nodes to be visited */
               current->visited &= 0x0;            /* 00000000 (mask that resets visit
                                                      order bits to 0) */
               if (current->right!=NULL) {
                  current->right->visited &= 0x0;  // 00000000 = Visit left subtree
                  toBeVisited.push(current->right);
               }
               break;
            default:
               break;
         }
      }
   }
   return retKeyList;
}

/* Checks whether a node exists; if so, it returns a pointer to
   its contents; otherwise, it returns a NULL pointer */
template <class elType, class keyType>
   elType* BST<elType, keyType>::findItem(const keyType& key) {
   elType *item;
   tree_node *current;
   current = root;
   while (current!=NULL) {
      if (current->key == key) {
         // Item has been found
         item = &(current->contents);
         return item;
      }
      if (current->key > key) {
         // Continue in the left subtree (if any)
         current = current->left;
      }
      else {
         // Continue in the right subtree (if any)
         current = current->right;
      }
   }
   return NULL;
}

// ******** Binary Search Tree (AVL tree) **********


#endif
