/********************************************************************/
/*    TORQUE (Types Of Relationships QUality Evaluation) toolkit    */
/*                                                                  */
/* Original file name: tpathExtract.h                               */
/*                                                                  */
/* This software by R.M.                                            */
/********************************************************************/

#ifndef TPATHEXTRACT_H
#define TPATHEXTRACT_H

#include "list.h"
#include "fileparsers.h"
#include "smartstructs.h"
#include "consoletools.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

/* The following are flags whose values correspond to the user-provided
   command line switches */
struct extraction_mode {
   extraction_mode() {
      sourceFile = NULL;
      outFile = NULL;
      cycleFile = NULL;
      prepFile = NULL;
      privFile = NULL;
      setFile = NULL;
      prepASnum = NULL;
      incompFile = NULL;
   }
   ~extraction_mode() {
      if (sourceFile != NULL) {
         delete sourceFile;
      }
      if (outFile != NULL) {
         delete outFile;
      }
      if (cycleFile != NULL) {
         delete cycleFile;
      }
      if (prepFile != NULL) {
         delete prepFile;
      }
      if (privFile != NULL) {
         delete privFile;
      }
      if (setFile != NULL) {
         delete setFile;
      }
      if (prepASnum != NULL) {
         delete prepASnum;
      }
      if (incompFile != NULL) {
         delete incompFile;
      }
   }
   char *sourceFile,             // Name of source file
        *outFile,                // Name of the file to write output in
        *cycleFile,              // Name of the file to save cycling paths in
        *prepFile,               // Name of the file to save paths with prepending in
        *privFile,               // Name of the file to save paths with private ASs in
        *setFile,                // Name of the file to save paths with AS sets in
        *incompFile;             // Name of the file to save incomplete paths in
   unsigned char prepending,     /* 0x0: keep prepending (i.e., do nothing about it)
                                    0x1: remove prepending
                                    0x2: delete paths with prepending from main output
                                    0x4: save paths with prepending to file
                                    0x8: show paths with prepending with a tag */
                 cycles,         /* 0x0: keep cycling paths (i.e., do nothing with them)
                                    0x1: remove cycles
                                    0x2: delete paths with cycles from main output
                                    0x4: save paths with cycles to file
                                    0x8: show paths with cycles with a tag */
                 privateAS,      /* 0x0: keep private ASs (i.e., do nothing with them)
                                    0x1: remove private ASs
                                    0x2: delete paths with private ASs from main output
                                    0x4: save paths with private ASs to file
                                    0x8: show paths with private ASs with a tag */
                 sets,           /* 0x1: remove AS sets (but keep the rest of the path)
                                    0x2: delete paths with AS sets from main output
                                    0x4: save paths with AS sets to file
                                    0x8: show paths with AS sets with a tag */
                 incomplete,     /* 0x0: keep incomplete paths
                                    0x2: delete incomplete paths
                                    0x4: save incomplete paths to file
                                    0x8: show incomplete paths with a tag */
                 prepAS;         /* 0x0: report paths as they appear
                                    0x1: always prepend a user provided AS number
                                    0x2: prepend user provided AS number only if needed */
   unsigned short int *prepASnum;// Number of AS to be prepended to each path
   // The following variables define the kind of output
   bool groupASpaths,            /* If this is true, then AS paths will be presented without
                                    repetitions, and with a number showing how many times
                                    each of them occurs in the BGP dump */
        showPrefixes,            /* Set to true if output shows, for each AS path, the
                                    prefix (or prefixes, depending on the value of
                                    `groupASpaths') that followed that path */
        alwaysPrefix,            /* If false, report NLRI as they appear in the BGP dump;
                                    when true, always report prefixes with their netmask */
        uniquePrefixes,          /* If true, prefixes will be reported without repetitions;
                                    this is only useful when grouping AS paths (i.e., when
                                    `groupASpaths' is true) */
        verbose,                 // If true, reports some statistics about processed paths
        toFile,                  // If true, output is written to a given file
        overwrite,               // If true, existing files will be automatically overwritten
        showHelp;                /* If true, displays information on how to use
                                    this program */
};

// A structure to hold path information
struct pathData {
   List<bgpParser::ASItem> ASpath;
   unsigned long int howMany;
   char *tags;
   List<IPAddress> network;
   List<IPNetmask*> netmask;
};

/* This function parses each option's mode and sets its second argument
   accordingly */
bool parseMode (char *string, unsigned char *mode);

/* Very similar to the previous function, but with specific
   options for prefixes settings */
bool parsePfxMode (char *sourceString, extraction_mode *mode);

/* The following function parses input arguments and puts results into
   `mode'; if an error occurs, it returns false and fills `error' with
   a printable error message; otherwise, it returns true */
bool parseArguments (int argc, char **argv, extraction_mode *mode, char **error);

// This function just prints a small header
void printHeader (FILE *);

// This functions shows instructions about program usage
void printHelp (FILE *);

/* This function displays a warning message for a file that is
   about to be overwritten, and asks user confirmation about it */
bool askOverwrite (char *fileName);

// Print a prefix, with network and (if either forced or needed) netmask
void printPrefix (FILE *, IPAddress, IPNetmask*, bool alwaysPrefix);

// This function prints an AS path
void printASpath (FILE *, List<bgpParser::ASItem>*);

// This function prepares an output file for writing into it
int openWriteFile(FILE **returnedStream, char *fileName, bool alwaysOverwrite);

/* This function takes as input two AS paths and returns true if
   they are the same path (including the flags) */
bool samePath (List<bgpParser::ASItem> &p1, List<bgpParser::ASItem> &p2);

// This function updates a map with the current AS path
void updateMap (BST<List<pathData*>*, unsigned short int> &pathSet,
                List<bgpParser::ASItem> &ASpath, char *tagStr, IPAddress network,
                IPNetmask *netmask, bool uniquePrefixes);

/* This function either writes to file or updates the
   corresponding path map (depending on the use of
   the --group option) */
void writeToFile (FILE *file, bool groupASpaths, char *tagStr,
                  bool showPrefixes, bool alwaysPrefix,
                  bool uniquePrefixes, bool tag,
                  IPAddress network, IPNetmask *netmask,
                  List<bgpParser::ASItem> &originalPath,
                  BST<List<pathData*>*, unsigned short int> &pathSet);

/* This function prints the contents of a path map to a file
   and at the same time releases memory that had been allocated
   for it */
void printMap (FILE *file, BST<List<pathData*>*, unsigned short int> &pathSet,
               bool tag, bool showPercentage, bool showPrefixes, bool alwaysPrefix);

// Cursor movement functions
void goToCol(int n);

#endif
